import asyncio
import websockets
from typing import Callable


class ServerConnection:
    def start(self, handler: Callable, host: str = "0.0.0.0", port: int = 1337):
        loop = asyncio.get_event_loop()
        loop.run_until_complete(websockets.serve(handler, host, port))
        loop.run_forever()



import inspect
import json
import websockets
import os
from typing import Dict, Optional, List

from quillion.utils.finder import RouteFinder
from .crypto import Crypto
from .messaging import Messaging
from .server import AssetServer, ServerConnection
from .router import Path
from ..pages.base import Page
from ..components import State


class Quillion:
    _instance = None

    def __init__(self):
        Quillion._instance = self
        self.callbacks: Dict[str, сallable] = {}
        self.current_page: Optional[Page] = None
        self.current_path: Optional[str] = None
        assets_host = os.environ.get("QUILLION_ASSET_HOST", "localhost")
        assets_port = os.environ.get("QUILLION_ASSET_PORT", "1338")
        self.assets_path = os.environ.get("QUILLION_ASSET_PATH", "")
        self.asset_server_url = f"http://{assets_host}:{assets_port}".rstrip("/")
        self.asset_server = AssetServer(assets_dir=self.assets_path)
        self.websocket = None
        self._state_instances: Dict[type, "State"] = {}
        self.style_tag_id = "quillion-dynamic-styles"
        self._current_rendering_page: Optional[Page] = None
        self.crypto = Crypto()
        self.messaging = Messaging(self)
        self.server_connection = ServerConnection()
        Path.init(self)
        self.external_css_files: List[str] = []
        self._css_cache: Dict[str, str] = {}

    def _load_css_file(self, css_file: str) -> str:
        from quillion_cli.debug.debugger import debugger

        if css_file in self._css_cache:
            return self._css_cache[css_file]

        with open(css_file, "r", encoding="utf-8") as f:
            content = f.read()
            self._css_cache[css_file] = content
            debugger.info(f"Loaded styles -> {css_file}")
            return content

    async def handler(self, websocket: websockets.WebSocketServerProtocol):
        from quillion_cli.debug.debugger import debugger

        self.websocket = websocket
        if self:
            connection_id = (
                f"{websocket.remote_address[0]}:{websocket.remote_address[1]}"
            )
            debugger.info(f"[{connection_id}] Received new connection")
        self._state_instances = {}
        initial_path = websocket.path
        try:
            public_key_message = await websocket.recv()
            data = json.loads(public_key_message)
            if await self.crypto.handle_key_exchange(websocket, data):
                await self.navigate(initial_path, websocket)
            else:
                return
            async for message in websocket:
                try:
                    data = json.loads(message)
                    inner_data = await self.crypto.decrypt_message(websocket, data)
                    if inner_data:
                        await self.messaging.process_inner_message(
                            websocket, inner_data
                        )
                except json.JSONDecodeError as e:
                    connection_id = (
                        f"{websocket.remote_address[0]}:{websocket.remote_address[1]}"
                    )
                    print(
                        f"[{connection_id}] json decode error: {e} - msg: {message}. not decrypted?"
                    )
                except Exception as e:
                    connection_id = (
                        f"{websocket.remote_address[0]}:{websocket.remote_address[1]}"
                    )
                    debugger.error(f"[{connection_id}] Error: {e}")
                    raise
        except Exception as e:
            connection_id = (
                f"{websocket.remote_address[0]}:{websocket.remote_address[1]}"
            )
            debugger.error(f"[{connection_id}] Error: {e}")
            raise
        finally:
            self._state_instances.clear()
            self.crypto.cleanup(websocket)

    async def navigate(
        self, path: str, websocket: websockets.WebSocketServerProtocol = None
    ):
        from quillion_cli.debug.debugger import debugger

        if path.startswith("http://") or path.startswith("https://"):
            content_message_for_encryption = {
                "action": "redirect",
                "url": path,
            }
            message_to_client = self.crypto.encrypt_response(
                websocket, content_message_for_encryption
            )
            await websocket.send(json.dumps(message_to_client))
            return

        page_cls, params, _ = RouteFinder.find_route(path)

        if page_cls and websocket:
            if not self.current_page or self.current_page.__class__ != page_cls:
                self.current_page = page_cls(params=params or {})
            self.current_path = path
            await self.render_current_page(websocket)
            connection_id = (
                f"{websocket.remote_address[0]}:{websocket.remote_address[1]}"
            )
            debugger.info(f"[{connection_id}] Redirected to: {path}")
        else:
            connection_id = (
                f"{websocket.remote_address[0]}:{websocket.remote_address[1]}"
            )
            debugger.error(f"[{connection_id}] Received unknown path: {path}")

    def redirect(self, path: str):
        if self.websocket:
            import asyncio

            asyncio.create_task(self.navigate(path, self.websocket))

    async def render_current_page(self, websocket: websockets.WebSocketServerProtocol):
        if not self.current_page or not websocket:
            return

        self._current_rendering_page = self.current_page
        self.current_page._rendered_component_keys.clear()

        for component_instance in self.current_page._component_instance_cache.values():
            component_instance._reset_hooks()

        try:
            root_element = self.current_page.render(**self.current_page.params)

            if inspect.isawaitable(root_element):
                root_element = await root_element

            from ..components.ui.base.container import Container

            if not isinstance(root_element, Container):
                root_element = Container(root_element)

            page_class_name = self.current_page.get_page_class_name()
            root_element.add_class(page_class_name)

            css_content = ""
            for css_file in self.external_css_files:
                css_content += self._load_css_file(css_file) + "\n"

            style_element = {
                "tag": "style",
                "attributes": {},
                "text": css_content,
                "children": [],
            }

            tree = root_element.to_dict(self)

            content = [style_element, tree]

            self.current_page._cleanup_old_component_instances()
            content_message_for_encryption = {
                "action": "render_page",
                "path": self.current_path,
                "content": content,
            }

            message_to_client = self.crypto.encrypt_response(
                websocket, content_message_for_encryption
            )
            await websocket.send(json.dumps(message_to_client))
        finally:
            self._current_rendering_page = None

    def css(self, files: List[str]):
        self.external_css_files.extend(files)
        return self

    def start(
        self, host="0.0.0.0", port=1337, assets_port=1338, assets_host="localhost"
    ):
        final_host = os.environ.get("QUILLION_HOST", host)
        final_port = int(os.environ.get("QUILLION_PORT", port))
        assets_port = int(os.environ.get("QUILLION_ASSET_PORT", assets_port))
        assets_host = os.environ.get("QUILLION_ASSET_HOST", assets_host)

        self.asset_server_url = f"http://{assets_host}:{assets_port}".rstrip("/")

        self.asset_server.start(host=assets_host, port=assets_port)

        self.server_connection.start(self.handler, final_host, final_port)



from typing import Optional, Callable
from ..element import Element


class Input(Element):
    def __init__(self, *children, class_name: Optional[str] = None, **kwargs):
        super().__init__("input", *children, class_name=class_name, **kwargs)


def input(*children, **kwargs):
    return Input(*children, **kwargs)



from typing import Optional
from ..element import MediaElement


class Canvas(MediaElement):
    def __init__(self, *children, class_name: Optional[str] = None, **kwargs):
        super().__init__("canvas", *children, class_name=class_name, **kwargs)


def canvas(*children, **kwargs):
    return Canvas(*children, **kwargs)



import asyncio
from typing import Optional, Dict


class Path:
    _app = None

    @classmethod
    def init(cls, app):
        cls._app = app

    @classmethod
    def navigate(cls, to: str, params: Optional[Dict[str, str]] = None):
        if not cls._app or not cls._app.websocket:
            return

        path = to.format(**params) if params else to
        asyncio.create_task(cls._app.navigate(path, cls._app.websocket))



from typing import Optional
from ..element import Element


class Text(Element):
    def __init__(self, *children, class_name: Optional[str] = None, **kwargs):
        super().__init__("p", *children, class_name=class_name, **kwargs)


def text(*children, **kwargs):
    return Text(*children, **kwargs)



from typing import Optional, Callable
from ..element import Element


class Button(Element):
    def __init__(self, *children, class_name: Optional[str] = None, **kwargs):
        super().__init__("button", *children, class_name=class_name, **kwargs)


def button(*children, **kwargs):
    return Button(*children, **kwargs)



from typing import Optional
from ..element import MediaElement


class Image(MediaElement):
    def __init__(self, *children, class_name: Optional[str] = None, **kwargs):
        super().__init__("img", *children, class_name=class_name, **kwargs)


def image(*children, **kwargs):
    return Image(*children, **kwargs)



from ..element import Element
from typing import List, Optional


class Container(Element):
    def __init__(self, *children, class_name: Optional[str] = None, **kwargs):
        super().__init__("div", class_name=class_name, **kwargs)

        for child in children:
            self.append(child)


def container(*children, **kwargs):
    return Container(*children, **kwargs)



from .element import Element, MediaElement
from .base.text import text
from .base.button import button
from .media.canvas import canvas
from .base.container import container
from .media.image import image
from .base.input import input



from typing import Dict, Tuple, Optional
import re
from .regex_parser import RegexParser
from ..pages.base import PageMeta


class RouteFinder:
    @staticmethod
    def _normalize_path(path: str) -> str:
        if not path or path == "/":
            return "/"

        path = path.strip("/")

        return f"/{path}"

    @staticmethod
    def find_route(path: str) -> Tuple[Optional[type], Optional[dict], float]:
        path = RouteFinder._normalize_path(path)
        path = path.strip()
        page_cls = None
        params = None
        max_priority = float("-inf")

        # check regex route
        for regex_pattern, (cls, priority) in PageMeta._regex_routes.items():
            match_params = RegexParser.extract_params(regex_pattern, path)
            if match_params is not None and priority > max_priority:
                page_cls, params, max_priority = cls, match_params, priority

        # check static routes
        if not page_cls:
            for route, (cls, priority) in PageMeta._registry.items():
                match_params = RegexParser.extract_params(cls._regex, path)
                if match_params is not None and priority > max_priority:
                    page_cls, params, max_priority = cls, match_params, priority

        # check dynamic routes
        if not page_cls:
            for route, (pattern, cls, priority) in PageMeta._dynamic_routes.items():
                match_params = RegexParser.extract_params(pattern, path)
                if match_params is not None and priority > max_priority:
                    page_cls, params, max_priority = cls, match_params, priority

        return page_cls, params, max_priority



import websockets
import inspect
import json
from typing import Dict, Any


class Messaging:
    def __init__(self, app):
        self.app = app

    async def process_inner_message(
        self, websocket: websockets.WebSocketServerProtocol, inner_data: Dict[str, Any]
    ):
        inner_action = inner_data.get("action")

        if inner_action == "callback":
            cb_id = inner_data.get("id")
            if cb_id in self.app.callbacks:
                cb = self.app.callbacks[cb_id]
                result = cb()
                if inspect.isawaitable(result):
                    await result
                await self.app.render_current_page(websocket)

        elif inner_action == "event_callback":
            cb_id = inner_data.get("id")
            event_data_str = inner_data.get("event_data", "{}")

            if cb_id in self.app.callbacks:
                cb = self.app.callbacks[cb_id]

                try:
                    event_data = json.loads(event_data_str) if event_data_str else {}
                except json.JSONDecodeError:
                    event_data = {}

                sig = inspect.signature(cb)
                if len(sig.parameters) > 0:
                    result = cb(event_data)
                else:
                    result = cb()

                if inspect.isawaitable(result):
                    await result
                await self.app.render_current_page(websocket)

        elif inner_action == "navigate":
            await self.app.navigate(inner_data.get("path", "/"), websocket)
        elif inner_action == "client_error":
            traceback = inner_data.get("error", "")
            print(
                f"\n[{websocket.remote_address[0]}:{websocket.remote_address[1]}] Error occurred"
            )
            print(traceback)
        else:
            print(
                f"[{websocket.remote_address[0]}:{websocket.remote_address[1]}] Unknown action: {inner_action}"
            )



import os
from typing import Optional, Dict, List, Any, Callable
import uuid
import re


class StyleProperty:
    def __init__(self, key: str, value):
        self._key = key
        self._value = value

    def to_css_properties_dict(self) -> Dict[str, str]:
        css_key = re.sub(r"([A-Z])", r"-\1", self._key).lower().replace("_", "-")
        css_value = str(self._value)
        return {css_key: css_value}


class Element:
    def __init__(
        self,
        tag: str,
        text: Optional[str] = None,
        event_handlers: Optional[Dict[str, Callable]] = None,
        inline_style_properties: Optional[Dict[str, str]] = None,
        classes: Optional[List[str]] = None,
        key: Optional[str] = None,
        class_name: Optional[str] = None,
        **kwargs: Any,
    ):
        self.tag = tag
        self.text = text
        self.event_handlers = event_handlers or {}
        self.children: List["Element"] = []
        self.attributes = {}
        self.inline_style_properties = inline_style_properties or {}
        self.css_classes = classes or []
        self.key = key
        self.style_properties: List["StyleProperty"] = []

        if class_name:
            self.css_classes.append(class_name)

        for prop_key, prop_value in kwargs.items():
            if prop_key.startswith("on_") and callable(prop_value):
                event_name = prop_key[3:]
                self.event_handlers[event_name] = prop_value
            elif prop_key == "style":
                if isinstance(prop_value, str):
                    styles = prop_value.split(";")
                    for style in styles:
                        if ":" in style:
                            css_key, css_value = style.split(":", 1)
                            self.inline_style_properties[css_key.strip()] = (
                                css_value.strip()
                            )
            else:
                self.style_properties.append(StyleProperty(prop_key, prop_value))

    def append(self, *children: "Element"):
        for child in children:
            self.children.append(child)
        return self

    def add_class(self, class_name: str):
        if class_name not in self.css_classes:
            self.css_classes.append(class_name)

    def add_event_handler(self, event_name: str, handler: Callable):
        self.event_handlers[event_name] = handler

    def set_attribute(self, name: str, value: Any):
        self.attributes[name] = value

    def to_dict(self, app) -> Dict[str, Any]:
        from ...components import CSS

        if isinstance(self, CSS):
            return self.to_dict(app)

        data = {
            "tag": self.tag,
            "attributes": self.attributes.copy(),
            "text": self.text,
            "children": [],
        }

        for event_name, handler in self.event_handlers.items():
            cb_id = str(uuid.uuid4())
            app.callbacks[cb_id] = handler
            data["attributes"][f"on{event_name}"] = cb_id

        all_styles = {}

        all_styles.update(self.inline_style_properties)

        for prop in self.style_properties:
            all_styles.update(prop.to_css_properties_dict())

        if all_styles:
            css_parts = [f"{k}: {v};" for k, v in all_styles.items()]
            data["attributes"]["style"] = " ".join(css_parts)

        if self.css_classes:
            if "class" in data["attributes"]:
                existing_class = data["attributes"]["class"]
                data["attributes"][
                    "class"
                ] = f"{existing_class} {' '.join(self.css_classes)}"
            else:
                data["attributes"]["class"] = " ".join(self.css_classes)

        if self.key:
            data["key"] = self.key

        for child in self.children:
            if isinstance(child, CSS):
                data["children"].append(child.to_dict(app))
            elif hasattr(child, "to_dict"):
                data["children"].append(child.to_dict(app))
            else:
                data["children"].append(child)

        return data


class MediaElement(Element):
    def __init__(
        self,
        tag: str,
        src: Optional[str] = None,
        event_handlers: Optional[Dict[str, Callable]] = None,
        inline_style_properties: Optional[Dict[str, str]] = None,
        classes: Optional[List[str]] = None,
        key: Optional[str] = None,
        class_name: Optional[str] = None,
        **kwargs: Any,
    ):
        super().__init__(
            tag,
            src,
            event_handlers,
            inline_style_properties,
            classes,
            key,
            class_name,
            **kwargs,
        )
        self.src = src

    def to_dict(self, app) -> Dict[str, Any]:
        result = super().to_dict(app)
        if self.src:
            if self.src.startswith("http://") or self.src.startswith("https://"):
                result["attributes"]["src"] = self.src
            else:
                result["attributes"][
                    "src"
                ] = f"{app.asset_server_url}/{self.src.replace(f'{app.assets_path}/', '')}"
        return result
